/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#include "spheres.h"
#include "tabusearch.h"
#include "statManager.h"
#include<iostream>
#include<iomanip>
#include<fstream>
#include<cstdlib>
#include<cstring>
#include<ctime>
#include<climits>
#include<cassert>
#include<algorithm>       //for simple min
using namespace std;

/*----------------------     Project-wide Global Variables      ---------------------*/
extern int**   edge;            //These global variables simply define the graph:
extern int     n;               //vtx no
extern int     m;               //edge no
extern int     k;               //cluster size
extern double**ww;              //weight of edge[i][j]: 1 if {i,j}\in V, 0 otherwise
extern int     objMaxPossible;  //maximum number of edges in a k-clique

/*--------------------       Global Data Used in main only      ---------------------*/
#define LOAD_RANDOM     INT_MAX
#define STD_LS          0       //Codes to identify the algorithm (here standard LS)
#define DGLS            1
#define STD_LS_RESTARTS 2
#define ST(tok)  #tok           
#define STR(tok) ST(tok)        //Classical transformation of macro to string

int srandseed         = 0;                                    
int algoVer           = STD_LS; //run standard LS by default, use -d for DGLS, -t for std ls with restarts
int itersElapsed      = 0;
int itersMax          = 1000000;//10^6 iters by default, modifiable using -i
int runs              = 10;     //we make statistics over 10 runs by default, modify via -m
int perturbStartSol;            //move away from start (opt) sol: change perturbStartSol vtx, 
                                //default k (full perturb), change with -p, if >n load random
statManager objVals;
statManager itersNeeded;
char* graphname;               
char* graphnameWithPath;               
/*------------------+-----------------------------------------------+--------------------------
                    |FUNCTIONS FOR INOUT, CONFIG DATA, CMD LINE ARGS|
                    +----------------------------------------------*/

//Reads a parameter (often the cluster size) from a file (config.txt)
int getParamFromFile(char*param, string filename)
{
    char token[100];
    ifstream fConfig(filename.c_str());
    if(fConfig.good())
        while(!fConfig.eof()){
            fConfig >> token;
            int minlen = strlen(token)<strlen(param)?strlen(token):strlen(param);
            if(!strncmp(param,token,minlen)){
                int toRet;
                fConfig >> toRet;
                fConfig.close();
                return toRet;
            }               
        }
    cerr<<"Please enter a line in "<<filename<<" for param/graph "<<param
        <<" I prefer to exit. Modify this code if you do not want to modify the file.\n";
    exit(EXIT_FAILURE);
}
//Linux kernel coding rules: macros resembling functions may be named in lower case
#define updateParam(cmdline_string,param,message)                                 \
    do{                                                                           \
        param = atoi(cmdline_string);                                             \
        clog<<message<<":"<<param<<endl;                                          \
        argc--;                                                                   \
        optionsFnd++;                                                             \
    }while(false);                                                                

void parseCmdLineArgs(int& argc, char** argv)
{
    if(argc==1) {
        cerr<<"Usage: ./main INSTANCEPATH "
              "[-v[erbose]] [-rRNDSEED] [-iMAXITER] [-mRUNS] [-d:useDGLS]"
              "[-pPERTURB STARTSOL SIZE] [-t:use STD LS with 5 restarts]\n";
        cerr<<"Default values:\n";
        cerr<<"MAXITER="<<::itersMax<<"\nRUNS="<<::runs<<"\nPERTURB=LOAD_RANDOM, i.e., "
              "start from rnd sol. Use -p15 to take a solution from INSTANCEPATH.sol "
              "and relocate 15 vertices!"<<endl;
        exit(EXIT_FAILURE);
    }
    //Command line arguments starting with '-'
    while(argv[argc-1][0]=='-') {
        int optionsFnd = 0;
        if( (argv[argc-1][0]=='-') && (!strncmp(argv[argc-1],"-v",2))) {
            cout<<"I'll be verbose, using clog."<<endl;
            clog.clear();
            argc--;
            optionsFnd++;
        }
        if(!strncmp(argv[argc-1],"-m",2))
            updateParam(argv[argc-1]+2,      ::runs,     "Number of runs/execs: ");
        if(!strncmp(argv[argc-1],"-r",2))
            updateParam(argv[argc-1]+2,      ::srandseed,"Using random seed: ");
        if(!strncmp(argv[argc-1],"-i",2))
            updateParam(argv[argc-1]+2,      ::itersMax, "Max nr of iterations ");
        if(!strncmp(argv[argc-1],"-d",2))
            updateParam(STR(DGLS),           ::algoVer,  "Use DGLS ");
        if(!strncmp(argv[argc-1],"-t",2))
            updateParam(STR(STD_LS_RESTARTS),::algoVer,  "Use standard Local Search with restarts");
        if(!strncmp(argv[argc-1],"-p",2))
            updateParam(argv[argc-1]+2,::perturbStartSol,"Clique bit values changed (perturbation size): ");

        if(optionsFnd==0) {
            cerr<<"There is an argument starting with '-' that I can not understand\n Accepted arguments:"
              "[-v[erbose]] [-rRNDSEED] [-iMAXITER] [-mRUNS] [-d:useDGLS]"
              "[-pPERTURB STARTSOL SIZE] [-t:use STD LS with restarts]\n";
            exit(EXIT_FAILURE);
        }
    }

    ::graphname         = new char[200];
    ::graphnameWithPath = new char[200];
    strcpy(::graphnameWithPath,argv[1]);
    strcpy(::graphname,argv[1]);
    while((strstr(::graphname,"/")!=NULL)||   //while the input file 
          (strstr(::graphname,"\\")!=NULL))   //has delimiters
                ::graphname = ::graphname+1;  //erase 1st char (not optimized, sorry)
}
void writeBestSol(int*sol){
    int nonEdges = 0;
    for(int i=0;i<(::n);i++)
        if(sol[i])
            for(int j=i+1;j<(::n);j++)
                if(sol[j])
                    if(!::edge[i][j])
                        nonEdges ++;
    char solOutfile[200];
    sprintf(solOutfile,"output/cliques/%s.k%d.obj%d.tm%ld.txt",graphname,k,nonEdges,time(NULL));
	ofstream fSol(solOutfile);
	if(fSol.fail()){
		cerr<<"Failed to open:"<<solOutfile<<" to write best solution."<<endl;
		return;
	}
	for(int jj=0;jj<(::n);jj++){
        if(sol[jj]==1)
		    fSol<<(jj+1)<<endl;//<<": "<<1<<endl;	
	}
	fSol<<endl;
    fSol.close(); 
}
/*-------------------------------------------------------------------------------------------------+
|                                                                                                  |
|                  Functions to launch (distance-guided) local search methods                      |
|                                                                                                  |
+-------------------------------------------------------------------------------------------------*/

void loadRandomSol(int* sol)
{
    for(int i=0;i<(::n);i++)
        sol[i] = 0;
    int i=0;
    while(i<k) {
        int rnd ;
        do {
            rnd = rand()%n;
        }while(sol[rnd]==1);
        sol[rnd]=1;
        i++;
    }
}

//test algos by perturbing an (optimal) input sol: change/shift noPerturbs bits
void perturbStartSolution(int* sol, int noPerturbs)
{
    int perturbsDone=0, i,j;
    for(j=n-1;j>=0;j--){                       //shift right
        if(perturbsDone==noPerturbs)
            goto writeRealSol;
        if(sol[j]==1){
            i = j;
            do{
                i=(i+1)%n;
                if(i==j){
                    cerr<<"It seems it is impossible to perturb solution"
                          "without keeping fixed some values. Par attention!";
                    goto writeRealSol;
                }
            }while(sol[i]!=0);
            sol[i] = 2;                        //shifted j
            clog<<j<<"shifted to "<<i<<"( 0-indexed)";
            perturbsDone++;
        } 
    }
    clog<<endl;
    if(perturbsDone!=noPerturbs){
        cerr<<"Can not apply "<<perturbStartSol<<"perturbations, k seems" 
               " less. Maybe reduce perturbStartSol. Exit now";
        exit(EXIT_FAILURE);
    }
    writeRealSol:                               //gotos ok for linux kernel rules
    for(j=j+1;j<n;j++)                          //all are shifted from j+1 on
        if(sol[j]==1)                           //and sol[j] are the initial
            sol[j]=0;                           //clique elements
    for(i=0;i<n;i++)
        if(sol[i]==2)
            sol[i]=1;
}

void loadSolFromFile(int*sol){
    ifstream input(strncat(graphnameWithPath,".sol",200));
    if(!input.good()){
        cerr<<"Can not find file to take input solution:"<<graphnameWithPath<<endl;
        exit(EXIT_FAILURE);
    }
    for(int i=0;i<(::n);i++)
        sol[i] = 0;
    clog<<"Start non-perturbed solution [1-indexed]:";
    for(int jj=0;jj<(::k);jj++){
        int idx;
        input>>idx;
        sol[idx-1] = 1;
        clog<<idx<<" ";
    }
    clog<<endl;
    input.close();
}
//Linux kernel coding rules: macros resembling functions may be named in lower case
#define initDataLS()                                                     \
    do{                                                                  \
        start      = new int[::n];                                       \
        bstCluster = new int[::n];                                       \
        if(::perturbStartSol==LOAD_RANDOM)                               \
            loadRandomSol(start);                                        \
        else{                                                            \
            loadSolFromFile(start);                                      \
            perturbStartSolution(start,::perturbStartSol);               \
        }                                                                \
        allocGlobalsTabuSearch();                                        \
        objMaxPossible=(::k)*(::k-1)/2;                                  \
    }while(false);

//Linux kernel coding rules: macros resembling functions may be named in lower case
#define logInfoFinishedLS()                                              \
    do{                                                                  \
        clog<<"best="<<bstObj<<" after iters="<<::itersElapsed<<endl;    \
        int nonEdges = objMaxPossible-bstObj;                            \
        clog<<"non-edges="<<nonEdges<<"\n===========================\n"; \
        cout<<"non-edges="<<nonEdges<<" itersDone="<<::itersElapsed<<endl;\
        if(nonEdges==0)                                                  \
            writeBestSol(bstCluster);                                    \
    }while(false);

void coutTexStats(string methodCode)
{
    char texGraph[200], remExtens[100];
    strcpy(remExtens,::graphname);
    if(remExtens[strlen(remExtens)-4]=='.')
        remExtens[strlen(remExtens)-4]='\0';
    sprintf(texGraph,"\\texttt{%s}",remExtens);
    cout<<setw(25)<<texGraph<<"& "<<::n<<"& "<<::k<<"& ";
    cout<<setw(7)<<methodCode<<"& ";
    if(perturbStartSol==LOAD_RANDOM)
        cout<<setw(4)<<"rnd "<<"& ";
    else
        cout<<setw(4)<<::perturbStartSol<<"& ";
    cout<<setw(2)<<::itersNeeded.getNrValues()<<"/"<<::runs<<"& ";
    cout<<::objVals.getStats()<<"& ";
    if(::itersNeeded.getNrValues()>0)
        cout<<::itersNeeded.getStats();
    else
        cout<<"&  --     & --     & --     & --     ";
    cout<<"\\\\%\n";
}

void stdLocalSearch(int maxruns)
{
    int *start, *bstCluster;
    initDataLS();

    for(int run=1;run<=maxruns;run++){
         srand(srandseed+run);                            //each run should call again srand
         cout<<"Starting std run using seed="<<srandseed+run<<":";
         double bstObj = tabuSearchSegment(bstCluster,::itersElapsed,              //output
                                           start,::itersMax,INFTY_RADIUS_NO_DISTS);//input

         logInfoFinishedLS();
         ::objVals.addValue(objMaxPossible-bstObj);
         if(bstObj==objMaxPossible)
            ::itersNeeded.addValue(::itersElapsed);
    }
    coutTexStats(" LS ");
    freeGlobalsTabuSearch();
}
double no_restarts_tmp = 0;
void stdLocalSearchWithRestarts(int maxruns, int noRestarts)
{
    int *start, *bstCluster;
    initDataLS();
    for(int run=1;run<=maxruns;run++){
        double bstObjAllRestarts = 0;
        double bstObj  = 0;                                             //bst one restart
        srand(srandseed+run);                                           //each run should call again srand
        cout<<"Starting std run using seed="<<srandseed+run<<":";
        ::itersElapsed = 0;
        for(int rst=1; (rst<=noRestarts) && (bstObj<objMaxPossible) ;rst++){
            no_restarts_tmp++;
            int itersOneRestart;
            bstObj = tabuSearchSegment(bstCluster,itersOneRestart,      //output
                                                 start,                 //input:
                                                 ::itersMax/noRestarts,
                                                 INFTY_RADIUS_NO_DISTS);
            ::itersElapsed += itersOneRestart;
            logInfoFinishedLS();
        }
        bstObjAllRestarts = max(bstObjAllRestarts,bstObj);
        ::objVals.addValue(objMaxPossible - bstObjAllRestarts);
        if(bstObjAllRestarts==objMaxPossible)
            ::itersNeeded.addValue(::itersElapsed);
    }
    cout<<"no_restarts_tmp:"<<no_restarts_tmp/maxruns<<endl;
    coutTexStats(" LS ");
    freeGlobalsTabuSearch();
}
void dgls(int maxruns)
{
    int *start, *bstCluster, *center, itersDoneLstRun, maxRadius = ::k*3/4 ;
    double bstObj;
    center = new int[::n];
    initDataLS();

    for(int run=1;run<=maxruns;run++){
        bstObj = 0; 
        srand(srandseed+run);                         //each run should call again srand
        cout<<"Starting dgls run using seed="<<srandseed+run<<":";
        cout.flush();
        itersElapsed = 0;
        enqueueCenter(start, INT_MAX);                      //start point has all priority
        do{
            dequeueMaxPriorCenter(center);
            no_restarts_tmp++;
            double bstObjSeg = tabuSearchSegment(bstCluster,itersDoneLstRun, //output values
                                       center,                               //input:
                                       min(10*::n,::itersMax-::itersElapsed),//
                                       maxRadius);                           //
            enqueueCenter(bstCluster,bstObjSeg);
#ifdef MDS_LOG
            arrows<<itersDoneLstRun<<" "<<idxGlob<<endl;
#endif
            bstObj = max(bstObj,bstObjSeg);
            ::itersElapsed = ::itersElapsed + itersDoneLstRun;
            clog<<"done iterations last run"<<itersDoneLstRun<<endl;
        } while( (::itersElapsed<(::itersMax)) && (bstObj<(::objMaxPossible)) );
#ifdef MDS_LOG
        printMat();
        exit(3);
#endif
        logInfoFinishedLS();
        ::objVals.addValue(objMaxPossible - bstObj);
        if(bstObj==objMaxPossible)
            ::itersNeeded.addValue(::itersElapsed);
        erasePrioritQueue();

    }
    coutTexStats("DGLS");
    cout<<"no_restarts_tmp:"<<no_restarts_tmp/maxruns<<endl;

    delete[] center;
    freeGlobalsTabuSearch();
}

int main(int argc, char**argv)
{
    clog.setstate(ios_base::failbit);          //clog disabled by default, use -v[erbose] to enable clog
    perturbStartSol = LOAD_RANDOM;
    parseCmdLineArgs(argc,argv);
    readGraph(argv[1]);                        //first argument = graph instance 
    ::k=getParamFromFile(graphname,"config.txt");
    clog<<"Using k="<<::k<<endl;
    srand(srandseed);                          //each run should call again 
                                               //srand to easily replicate experiments
    cout<<"Note: unless you specified -v[erbose], I will only print the number of non-edges\n"
          "at the end of each run and a final tex table with statistical results\n"
          "By default I run a standard LS, use -d[gls] for Distance Guided LS\n"<<endl;
    if(::algoVer == DGLS)
        dgls(::runs);
    if(::algoVer == STD_LS)
        stdLocalSearch(::runs);
    if(::algoVer == STD_LS_RESTARTS)
        stdLocalSearchWithRestarts(::runs, 5 ); //5 restarts
}
